@@ -1,5 +1,6 @@
 /*
- *  Copyright (c) 2008 Mike Massonnet <mmassonnet@xfce.org>
+ *  Copyright (c) 2008-2009 Mike Massonnet <mmassonnet@xfce.org>
+ *  Copyright (c) 2009 Vincent Legout <vincent@xfce.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -9,11 +10,11 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -27,10 +28,13 @@
 #include "playlist.h"
 #include "preferences.h"
 #include "mpdclient.h"
+#include "dbbrowser.h"
+#include "extended-interface.h"
+#include "song-dialog.h"
 
 #define BORDER 4
 
-#define XFMPC_PLAYLIST_GET_PRIVATE(o) \
+#define GET_PRIVATE(o) \
     (G_TYPE_INSTANCE_GET_PRIVATE ((o), XFMPC_TYPE_PLAYLIST, XfmpcPlaylistPrivate))
 
 
@@ -60,12 +64,17 @@ static gboolean         visible_func_filter_tree                (GtkTreeModel *f
                                                                  GtkTreeIter *iter,
                                                                  XfmpcPlaylist *playlist);
 
+void                    cb_browse_selection                     (XfmpcPlaylist *playlist);
+void                    cb_info_selection                       (XfmpcPlaylist *playlist);
+
 
 
 /* List store identifiers */
 enum
 {
   COLUMN_ID,
+  COLUMN_FILENAME,
+  COLUMN_POSITION,
   COLUMN_SONG,
   COLUMN_LENGTH,
   COLUMN_WEIGHT,
@@ -81,22 +90,23 @@ struct _XfmpcPlaylistClass
 
 struct _XfmpcPlaylist
 {
-  GtkVBox               parent;
-  XfmpcPlaylistPrivate *priv;
-  XfmpcPreferences     *preferences;
-  XfmpcMpdclient       *mpdclient;
+  GtkVBox                   parent;
+  XfmpcPreferences         *preferences;
+  XfmpcMpdclient           *mpdclient;
+  /*<private>*/
+  XfmpcPlaylistPrivate     *priv;
 };
 
 struct _XfmpcPlaylistPrivate
 {
-  GtkWidget            *treeview;
-  GtkTreeModelFilter   *filter;
-  GtkListStore         *store;
-  GtkWidget            *filter_entry;
-  GtkWidget            *menu;
+  GtkWidget                *treeview;
+  GtkTreeModelFilter       *filter;
+  GtkListStore             *store;
+  GtkWidget                *filter_entry;
+  GtkWidget                *menu;
 
-  gint                  current;
-  gboolean              autocenter;
+  gint                      current;
+  gboolean                  autocenter;
 };
 
 
@@ -150,7 +160,7 @@ xfmpc_playlist_class_init (XfmpcPlaylistClass *klass)
 static void
 xfmpc_playlist_init (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = playlist->priv = GET_PRIVATE (playlist);
 
   playlist->preferences = xfmpc_preferences_get ();
   playlist->mpdclient = xfmpc_mpdclient_get ();
@@ -161,6 +171,8 @@ xfmpc_playlist_init (XfmpcPlaylist *playlist)
 
   /* === Tree model === */
   priv->store = gtk_list_store_new (N_COLUMNS,
+                                    G_TYPE_INT,
+                                    G_TYPE_STRING,
                                     G_TYPE_INT,
                                     G_TYPE_STRING,
                                     G_TYPE_STRING,
@@ -184,8 +196,19 @@ xfmpc_playlist_init (XfmpcPlaylist *playlist)
   gtk_tree_view_set_model (GTK_TREE_VIEW (priv->treeview), GTK_TREE_MODEL (priv->filter));
   g_object_unref (priv->filter);
 
-  /* Column "artist - title" */
+  /* Column "position" */
   GtkCellRenderer *cell = gtk_cell_renderer_text_new ();
+  g_object_set (G_OBJECT (cell),
+                "xalign", 1.0,
+                NULL);
+  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (priv->treeview),
+                                               -1, "Length", cell,
+                                               "text", COLUMN_POSITION,
+                                               "weight", COLUMN_WEIGHT,
+                                               NULL);
+
+  /* Column "artist - title" */
+  cell = gtk_cell_renderer_text_new ();
   g_object_set (G_OBJECT (cell),
                 "ellipsize", PANGO_ELLIPSIZE_END,
                 NULL);
@@ -224,6 +247,18 @@ xfmpc_playlist_init (XfmpcPlaylist *playlist)
   g_signal_connect_swapped (mi, "activate",
                             G_CALLBACK (xfmpc_playlist_delete_selection), playlist);
 
+  mi = gtk_image_menu_item_new_with_mnemonic (_("_Browse"));
+  gtk_menu_shell_append (GTK_MENU_SHELL (priv->menu), mi);
+  g_signal_connect_swapped (mi, "activate",
+                            G_CALLBACK (cb_browse_selection), playlist);
+  GtkWidget *image = gtk_image_new_from_stock (GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
+  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mi), image);
+
+  mi = gtk_image_menu_item_new_from_stock (GTK_STOCK_INFO, NULL);
+  gtk_menu_shell_append (GTK_MENU_SHELL (priv->menu), mi);
+  g_signal_connect_swapped (mi, "activate",
+                            G_CALLBACK (cb_info_selection), playlist);
+
   gtk_widget_show_all (priv->menu);
 
   /* === Filter entry === */
@@ -255,6 +290,12 @@ xfmpc_playlist_init (XfmpcPlaylist *playlist)
                             G_CALLBACK (cb_filter_entry_key_released), playlist);
   g_signal_connect_swapped (priv->filter_entry, "changed",
                             G_CALLBACK (cb_filter_entry_changed), playlist);
+
+  /* Preferences */
+  g_signal_connect_swapped (playlist->preferences, "notify::song-format",
+                            G_CALLBACK (cb_playlist_changed), playlist);
+  g_signal_connect_swapped (playlist->preferences, "notify::song-format-custom",
+                            G_CALLBACK (cb_playlist_changed), playlist);
 }
 
 static void
@@ -282,15 +323,19 @@ xfmpc_playlist_new ()
 void
 xfmpc_playlist_append (XfmpcPlaylist *playlist,
                        gint id,
+                       gint pos,
+                       gchar *filename,
                        gchar *song,
                        gchar *length)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   GtkTreeIter           iter;
 
   gtk_list_store_append (priv->store, &iter);
   gtk_list_store_set (priv->store, &iter,
                       COLUMN_ID, id,
+                      COLUMN_FILENAME, filename,
+                      COLUMN_POSITION, pos + 1,
                       COLUMN_SONG, song,
                       COLUMN_LENGTH, length,
                       COLUMN_WEIGHT, PANGO_WEIGHT_NORMAL,
@@ -300,15 +345,14 @@ xfmpc_playlist_append (XfmpcPlaylist *playlist,
 void
 xfmpc_playlist_clear (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
-
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   gtk_list_store_clear (priv->store);
 }
 
 void
 xfmpc_playlist_delete_selection (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   GtkTreeModel         *store = GTK_TREE_MODEL (priv->store);
   GtkTreeIter           iter;
   GList                *list;
@@ -338,7 +382,7 @@ void
 xfmpc_playlist_select_row (XfmpcPlaylist *playlist,
                            gint i)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
 
   if (gtk_tree_model_iter_n_children (GTK_TREE_MODEL (priv->filter), NULL) == 0)
     return;
@@ -352,7 +396,7 @@ xfmpc_playlist_select_row (XfmpcPlaylist *playlist,
 void
 xfmpc_playlist_refresh_current_song (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
 
   /* Remove the bold from the "last" current song */
   GtkTreeIter iter;
@@ -376,7 +420,7 @@ xfmpc_playlist_refresh_current_song (XfmpcPlaylist *playlist)
 static void
 cb_song_changed (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
 
   xfmpc_playlist_refresh_current_song (playlist);
 
@@ -388,16 +432,17 @@ cb_song_changed (XfmpcPlaylist *playlist)
 static void
 cb_playlist_changed (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
-  gchar                *song, *length;
-  gint                  id, current;
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
+  gchar                *filename, *song, *length;
+  gint                  id, pos, current;
 
   current = xfmpc_mpdclient_get_id (playlist->mpdclient);
 
   xfmpc_playlist_clear (playlist);
-  while (xfmpc_mpdclient_playlist_read (playlist->mpdclient, &id, &song, &length))
+  while (xfmpc_mpdclient_playlist_read (playlist->mpdclient, &id, &pos, &filename, &song, &length))
     {
-      xfmpc_playlist_append (playlist, id, song, length);
+      xfmpc_playlist_append (playlist, id, pos, filename, song, length);
+      g_free (filename);
       g_free (song);
       g_free (length);
     }
@@ -414,7 +459,7 @@ cb_row_activated (XfmpcPlaylist *playlist,
                   GtkTreePath *path,
                   GtkTreeViewColumn *column)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   GtkTreeIter           iter;
   gint                  id;
 
@@ -451,7 +496,7 @@ static gboolean
 cb_button_pressed (XfmpcPlaylist *playlist,
                    GdkEventButton *event)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   GtkTreeSelection     *selection;
   GtkTreePath          *path;
   
@@ -480,15 +525,13 @@ static gboolean
 cb_popup_menu (XfmpcPlaylist *playlist)
 {
   popup_menu (playlist);
-
   return TRUE;
 }
 
 static void
 popup_menu (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
-
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   gtk_menu_popup (GTK_MENU (priv->menu),
                   NULL, NULL,
                   NULL, NULL,
@@ -501,7 +544,7 @@ popup_menu (XfmpcPlaylist *playlist)
 static void
 cb_filter_entry_activated (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   GtkTreeModel         *model = GTK_TREE_MODEL (priv->filter);
   GList                *list;
 
@@ -521,7 +564,7 @@ static gboolean
 cb_filter_entry_key_released (XfmpcPlaylist *playlist,
                               GdkEventKey *event)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
 
   if (event->type != GDK_KEY_RELEASE)
     return FALSE;
@@ -546,8 +589,7 @@ cb_filter_entry_key_released (XfmpcPlaylist *playlist,
 static void
 cb_filter_entry_changed (XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
-
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   gtk_tree_model_filter_refilter (priv->filter);
 }
 
@@ -556,7 +598,7 @@ visible_func_filter_tree (GtkTreeModel *filter,
                           GtkTreeIter *iter,
                           XfmpcPlaylist *playlist)
 {
-  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST_GET_PRIVATE (playlist);
+  XfmpcPlaylistPrivate *priv = XFMPC_PLAYLIST (playlist)->priv;
   gchar                *song, *song_tmp;
   gchar                *search;
   gboolean              result = TRUE;
@@ -586,3 +628,72 @@ visible_func_filter_tree (GtkTreeModel *filter,
   return result;
 }
 
+void
+cb_browse_selection (XfmpcPlaylist *playlist)
+{
+  XfmpcPlaylistPrivate   *priv = XFMPC_PLAYLIST (playlist)->priv;
+  XfmpcDbbrowser         *dbbrowser;
+  XfmpcExtendedInterface *extended_interface;
+  GtkTreeModel           *store = GTK_TREE_MODEL (priv->store);
+  GtkTreeSelection       *selection;
+  GtkTreeIter             iter;
+  GList                  *list;
+  gchar                  *filename, *dir;
+
+  dbbrowser = g_object_get_data (G_OBJECT (playlist), "XfmpcDbbrowser");
+  extended_interface = g_object_get_data (G_OBJECT (playlist), "XfmpcExtendedInterface");
+
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview));
+  if (gtk_tree_selection_count_selected_rows (selection) > 1)
+    return;
+
+  list = gtk_tree_selection_get_selected_rows (selection, &store);
+  if (gtk_tree_model_get_iter (store, &iter, list->data))
+    {
+      gtk_tree_model_get (store, &iter,
+                          COLUMN_FILENAME, &filename,
+                          -1);
+
+      dir = g_path_get_dirname (filename);
+      xfmpc_dbbrowser_set_wdir (dbbrowser, dir);
+      xfmpc_dbbrowser_reload (dbbrowser);
+      xfmpc_extended_interface_set_active (extended_interface, XFMPC_EXTENDED_INTERFACE_DBBROWSER);
+
+      g_free (filename);
+      g_free (dir);
+    }
+
+  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
+  g_list_free (list);
+}
+
+void
+cb_info_selection (XfmpcPlaylist *playlist)
+{
+  XfmpcPlaylistPrivate   *priv = XFMPC_PLAYLIST (playlist)->priv;
+
+  GtkTreeModel           *store = GTK_TREE_MODEL (priv->store);
+  GtkTreeSelection       *selection;
+  GtkTreeIter             iter;
+  GList                  *list;
+  gint                    id;
+
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview));
+  if (gtk_tree_selection_count_selected_rows (selection) > 1)
+    return;
+
+  list = gtk_tree_selection_get_selected_rows (selection, &store);
+  if (gtk_tree_model_get_iter (store, &iter, list->data))
+    {
+      gtk_tree_model_get (store, &iter,
+                          COLUMN_ID, &id,
+                          -1);
+
+      GtkWidget *dialog = xfmpc_song_dialog_new (id);
+      gtk_widget_show (dialog);
+    }
+
+  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
+  g_list_free (list);
+}
+
